// a) 都道府県郡市区町村名csv : [地方公共団体コード|false]
// b) 地方公共団体コード : {"町名":"丁目数値のcsv"}
const dictionary = require("./lib/dictionary");
const util = require("./lib/util");

// 都道府県市区町村を一意に特定できるフレーズの集合
const upper = (function() {
  const src = {};
  Object.keys(dictionary).filter(key => key.match(/[都道府県郡市区町村]$/)).forEach(key => {
    const tokens = key.split(",");
    const labels = [tokens.pop()];
    while (tokens.length > 0) {
      const tail = tokens.pop();
      labels.forEach(label => {
        labels.push(tail + label);
      });
    }
    if (key.match(/郡$/)) key = key.replace(/,.*$/, "");
    labels.forEach(label => {
      if (src[label] === undefined) src[label] = [];
      if (src[label].indexOf(key) == -1) src[label].push(key);
    });
  });
  return src;
})();

// 直下に町名を持つようなトークンの集合
const all = Object.keys(dictionary).filter(token => dictionary[dictionary[token]] !== undefined);

function log(target, type, message) {
  if (type === "INFO") {
    const added = ["都道府県", "市区町村", "区", "町名", "丁目", "番地", "号"].filter(a => target[a] !== undefined);
    if (added.length === 0) message = "プロパティは追加されませんでした";
    else {
      const fixed = ["都道府県", "市区町村", "区", "町名"].filter(a => target[a] !== undefined && target["表記"].replace(/[ 　]/g, "").indexOf(target[a]) === -1);
      if (fixed.length === 0) {
        message = `プロパティ(${added.join("|")}) が追加されました`;
      } else if (fixed.length === 1 && fixed[0] === "都道府県") {
        message = `プロパティ(${added.join("|")}) が追加され、都道府県の値が補完されました`;
      } else {
        message = `プロパティ(${added.join("|")}) が追加され、そのうち (${fixed.join("|")}) の値が補完されました。`;
        type = "WARNING";
      }
    }
  }
  const now = new Date();
  now.setHours(now.getHours() + 9);
  const content = {
    "@type": "hj:更新履歴型",
    "hj:更新者": {
      "@type": "実体型",
      "表記": "IMI住所変換コンポーネント"
    },
    "種別": type,
    "説明": message,
    "日時": {
      "@type": "日時型",
      "標準型日時": now.toISOString().replace(/Z$/, "+09:00")
    }
  };
  if (target["メタデータ"] === undefined) target["メタデータ"] = content;
  else if (Array.isArray(target["メタデータ"])) target["メタデータ"].push(content);
  else target["メタデータ"] = [target["メタデータ"], content];
}

function main(target) {

  // 表記の前後の空白はトリムする
  let label = (target["表記"]["@value"] || target["表記"]).replace(/[ 　]/g, "");

  // カタカナまたは特定漢字名で始まる場合は除外
  if (label.match(/^([\u30a0-\u30ff]+|大韓民国|中華人民共和国|香港)/)) {
    log(target, "WARNING", "日本国内の住所ではないようです");
    return;
  }

  // 都道府県郡市区町村キーワードがなにもない
  if (!label.match(/^(北海道|東京都|.+[府県市区町村])/)) {
    log(target, "WARNING", "日本国内の住所ではないようです");
    return;
  }

  // 表記の前後を削って都道府県市区町村名と合致するケースをピックアップ
  let candidate = [];
  for (let start = 0; start < label.length - 1; start++) {
    for (let end = label.length; start < end; end--) {
      const head = label.substring(start, end);
      const tail = label.substring(end);
      if (upper[head]) {
        upper[head].forEach(token => {
          if (candidate.find(x => x.token === token && x.tail === tail)) return;
          candidate.push({
            token: token,
            tail: tail,
            head: head,
            start: start
          });
        });
        break;
      }
    }
  }

  let fallbackPrefecture = "";

  // 都道府県と市区町村が混在する場合は市区町村だけを採用する
  if (candidate.find(a => a.token.match(/[市区町村]$/))) {
    candidate = candidate.filter(a => {
      if (a.token.match(/[都道府県]$/)) fallbackPrefecture = a.token;
      return a.token.match(/[市区町村]$/);
    });
  }

  // マッチした位置が先頭のほうのものを採用する（市区町村の町村名と下位町名がバッティングすることがあるので）
  const minStart = candidate.reduce((acc, cur) => Math.min(acc, cur.start), label.length);
  candidate = candidate.filter(x => x.start === minStart);

  // マッチした長さが最長のものを採用する
  const maxHeadLength = candidate.reduce((acc, cur) => Math.max(acc, cur.head.length), 0);
  candidate = candidate.filter(x => x.head.length === maxHeadLength);

  // 都道府県・政令指定都市の市までがわかっている場合
  if (candidate.length === 1) {
    const a = candidate[0];
    const code = dictionary[a.token];
    if (code !== false && dictionary[code] !== undefined && Object.keys(dictionary[code]).length === 0) {
      const b = all.filter(token => token.indexOf(a.token + ",") === 0);
      if (b.length > 0)
        candidate = b.map(token => {
          return {
            token: token,
            tail: a.tail
          };
        });
    }
  }

  /*
    // 候補が都道府県の場合には町名探索の範囲を都道府県配下の全市区町村とする
    if (candidate.length !== 0 && candidate[0].token.match(/[都道府県]$/)) {
      const pref = candidate.pop();
      all.filter(token => token.indexOf(pref.token + ",") === 0).forEach(token => {
        candidate.push({
          token: token,
          tail: pref.tail
        });
      });
    }*/


  // 候補が空の場合は町名探索の範囲を全市区町村とする
  if (candidate.length === 0) {
    all.forEach(token => {
      candidate.push({
        token: token,
        tail: label
      });
    });
  }

  // 候補を絞り込む
  //  let live = candidate.length === 1 ? candidate : candidate.filter(x => {
  let live = candidate.filter(x => {
    const code = dictionary[x.token];
    // 地方公共団体コードがない場合、指定された地方公共団体コード配下の町名オブジェクトがない場合は除外
    if (code === undefined || dictionary[code] === undefined) return false;

    // 末尾文字列の候補、揺らぎがある場合は複数候補をたてる
    let tails = [x.tail];
    // ケヶ問題
    if (x.tail.indexOf("ヶ") !== -1) tails.push(x.tail.replace("ヶ", "ケ"));
    if (x.tail.indexOf("ケ") !== -1) tails.push(x.tail.replace("ケ", "ヶ"));
    // 北海道の００条００丁目
    if (x.tail.match(/^([^0-9０-９]+)([0-9０-９]+)(条.*)$/)) {
      const param = [RegExp.$1, RegExp.$2, RegExp.$3];
      param[1] = util.h2k(util.z2h(param[1]));
      tails.push(param.join(""));
    } else if (x.tail.match(/^([0-9０-９]+)(条.*)$/)) {
      const param = [RegExp.$1, RegExp.$2];
      param[0] = util.h2k(util.z2h(param[0]));
      tails.push(param.join(""));
    }
    // 一番町
    if (x.tail.match(/^([^0-9０-９]+)([0-9０-９]+)(番町.*)$/)) {
      const param = [RegExp.$1, RegExp.$2, RegExp.$3];
      param[1] = util.h2k(util.z2h(param[1]));
      tails.push(param.join(""));
    } else if (x.tail.match(/^([0-9０-９]+)(番町.*)$/)) {
      const param = [RegExp.$1, RegExp.$2];
      param[0] = util.h2k(util.z2h(param[0]));
      tails.push(param.join(""));
    }

    //町名が長い順にソート
    let names = Object.keys(dictionary[code]);
    names.sort((a, b) => b.length - a.length);

    // 互いに無関係になる name,tail は取り除く
    tails = tails.filter(a => names.find(b => a.indexOf(b) !== -1));
    names = names.filter(a => tails.find(b => b.indexOf(a) !== -1));

    // name か tail が空だった場合は終了
    if (tails.length === 0 || names.length === 0) return false;

    // tail は一つになっているはずなのでこちらで
    let tail = tails.find(a => a.indexOf(x.tail) !== -1) || tails.pop();
    for (let i = 0; i < tail.length; i++) {
      const body = tail.substring(i);
      const name = names.find(a => body.indexOf(a) === 0);
      if (name) {
        const head = tail.substring(0, tail.indexOf(name));
        const rest = tail.substring(tail.indexOf(name) + name.length);
        // 京都市の特殊住所対応
        if (x.token.match(/京都市,.*区/) && rest.match(/(上る|上ル|下る|下ル)/)) {
          continue;
        }
        x.chomei = name;
        x.name = ("" + x.tail);
        if (head.length > 0) x.name = x.name.substring(x.name.indexOf(head) + head.length);
        if (rest.length > 0) x.name = x.name.substring(0, x.name.lastIndexOf(rest));
        x.tail = rest;
        return true;
      }
    }
    return false;
  });

  // 町名が見つからなかった場合にはここで終了
  if (live.length === 0) {
    // もっともヒット数の多いトークンだけを残す
    let tokens = candidate.map(c => c.token);
    let max = 0;
    tokens.forEach(token => {
      const score = token.split(",").filter(a => label.indexOf(a) !== -1).length;
      max = Math.max(score, max);
    });
    tokens = tokens.filter(token => {
      const score = token.split(",").filter(a => label.indexOf(a) !== -1).length;
      return score === max;
    });

    // トークンの最長の共通部分
    let token = tokens.pop();
    tokens.forEach(a => {
      let tmp = a.split(",");
      while (tmp.length > 0 && token.indexOf(tmp.join(",")) !== 0)
        tmp.pop();
      token = tmp.join(",");
    });
    if (token.length > 0) {
      token.split(",").forEach(a => {
        if (a.match(/[都道府県]$/)) target["都道府県"] = a;
        else if (a.match(/[市町村]$/)) target["市区町村"] = a;
        else if (a.match(/[区]$/)) {
          if (target["市区町村"] === undefined) target["市区町村"] = a;
          else target["区"] = a;
        }
      });

      if (dictionary[token] === false) log(target, "WARNING", "過去に存在した市区町村");
      else log(target, "INFO");
    } else if (fallbackPrefecture !== "") {
      target["都道府県"] = fallbackPrefecture;
      log(target, "INFO");
    } else {
      log(target, "WARNING", "都道府県を特定できませんでした");
    }
    return;
  }

  // 候補町名の中で最長のものを残す
  if (live.length > 1) {
    const maxNameLength = live.reduce((acc, cur) => Math.max(acc, cur.name.length), 0);
    live = live.filter(a => a.name.length === maxNameLength);
  }

  // 候補町名のヒット数が最大のものを残す
  if (live.length > 1) {
    let max = 0;
    live.forEach(a => {
      max = Math.max(max, a.token.split(",").filter(b => label.indexOf(b) !== -1).length);
    });
    live = live.filter(a => max === a.token.split(",").filter(b => label.indexOf(b) !== -1).length);
  }
  // 候補が唯一でない場合にはここで終了
  if (live.length > 1) {
    const pref = [];
    live.forEach(p => {
      const x = p.token.split(",").shift();
      if (pref.indexOf(x) === -1) pref.push(x);
    });
    if (pref.length === 1) {
      target["都道府県"] = pref[0];
    }

    log(target, "WARNING", "複数の候補があり、特定できませんでした [" + live.map(a => a.token.replace(/,/g, "") + a.name).join("|") + "]");
    return;
  }

  const found = live[0];

  // トークンをもとに都道府県市区町村を補完
  found.token.split(",").forEach(a => {
    if (a.match(/[都道府県]$/)) target["都道府県"] = a;
    else if (a.match(/[市町村]$/)) target["市区町村"] = a;
    else if (a.match(/[区]$/)) {
      if (target["市区町村"] === undefined) target["市区町村"] = a;
      else target["区"] = a;
    }
  });

  // 市区町村の補完ができなかった場合は終了
  if (target["市区町村"] === undefined) {
    log(target, "WARNING", "住所から市区町村を特定できませんでした");
    return;
  }

  // 過去に存在した自治体の場合には終了
  if (dictionary[found.token] === false) {
    log(target, "WARNING", "過去に存在した市区町村");
    return;
  }

  target["町名"] = found.name;

  let foot = found.tail;

  // 桁数の多い番地で全角桁区切りの用例がある
  foot = foot.replace(/，/g, "");

  // 残存文字列がない場合には終了
  if (foot.length === 0) {
    log(target, "INFO");
    return;
  }



  // 残存文字列の先頭が数字でない場合には終了
  if (foot.match(/^[^0123456789０１２３４５６７８９〇一二三四五六七八九十]+/)) {
    log(target, "INFO");
    return;
  }

  // 丁目
  // 丁目＋番地
  // 丁目＋番地＋号
  // 番地
  // 番地＋号

  const bangou = foot.split(/[^0123456789０１２３４５６７８９〇一二三四五六七八九十]+/).map(a => {
    a = util.z2h(a);
    a = util.k2h(a);
    return a;
  });

  if (!foot.match(/^[0123456789０１２３４５６７８９〇一二三四五六七八九十]+(番地|番)/)) {

    const chome = dictionary[dictionary[found.token]][found.chomei].split(",");

    // 辞書に定義された丁目がある場合
    if (chome.indexOf(bangou[0]) !== -1) {
      target["丁目"] = bangou[0];
      if (bangou[1]) {
        target["番地"] = bangou[1];
        if (bangou[2]) {
          target["号"] = bangou[2];
        }
      }
      log(target, "INFO");
      return;
    }

    // 丁目が必須のパターン
    if (bangou[0] < 45 && chome.find(a => a === "") === undefined) {
      if (bangou[0]) {
        target["丁目"] = bangou[0];
        if (bangou[1]) {
          target["番地"] = bangou[1];
          if (bangou[2]) {
            target["号"] = bangou[2];
          }
        }
      }
      log(target, "WARNING", `辞書に登録のない丁目です [${target["町名"]}${bangou[0]}丁目]`);
      return;
    }

  }

  if (foot.indexOf("合地") !== -1 || foot.indexOf("合併地") !== -1) {
    // 番地が複数になるのでなにもしない
  } else if (foot.match(/^[0123456789０１２３４５６７８９〇一二三四五六七八九十]+(番地の|番地|番|−|-|ー|‐|―|の|？)[0123456789０１２３４５６７８９〇一二三四五六七八九十]+/)) {
    // 番地＋デリミタ＋号（＋任意の文字列）
    target["番地"] = bangou[0];
    target["号"] = bangou[1];
  } else if (foot.match(/^[0123456789０１２３４５６７８９〇一二三四五六七八九十]+$/)) {
    // 番地だけ
    target["番地"] = bangou[0];
  } else if (foot.match(/^[0123456789０１２３４５６７８９〇一二三四五六七八九十]+(番地|番)/)) {
    // 数字または漢字で表現された〇〇番地＋ビル名
    target["番地"] = bangou[0];
  } else if (foot.match(/^[0123456789０１２３４５６７８９〇一二三四五六七八九十]+(丁目|字|部|線|号|地割|街区|区|丁|組|階)/)) {
    // 番地・号とは違う番号体系が導入されているのでなにもしない
  } else if (foot.match(/^[0123456789０１２３４５６７８９]+[^0123456789０１２３４５６７８９]/)) {
    // 数字だけで表現された番地＋ビル名のパターン
    target["番地"] = bangou[0];
  } else {
    // ここにくるのは漢字始まりのケース、下位地名とかんがえてなにもしない
  }
  log(target, "INFO");
}

const isTarget = (a) => {
  if (["住所型", "hj:住所型"].indexOf(a["@type"]) === -1) return false;
  return (a["表記"] !== undefined);
};

module.exports = function(source) {
  let count = 0;
  const dig = function(a) {
    if (Array.isArray(a)) {
      a.forEach(b => {
        dig(b);
      });
    } else if (typeof a === 'object') {
      if (isTarget(a)) {
        main(a);
        count++;
      } else {
        Object.keys(a).forEach(key => {
          dig(a[key]);
        });
      }
    }
    return a;
  };

  const target = dig(JSON.parse(JSON.stringify(source)));

  // JSON-LD Context の付与
  const context = [
    "https://imi.go.jp/ns/core/context.jsonld",
    "http://hojin-info.go.jp/ns/domain/biz/1/context.jsonld"
  ];
  if (target["@context"] === undefined) {
    target["@context"] = context;
  } else if (Array.isArray(target["@context"])) {
    context.forEach(a => {
      if (target["@context"].indexOf(a) === -1) target["@context"].push(a);
    });
  } else if (typeof target["@context"] === 'string') {
    target["@context"] = [target["@context"]];
    context.forEach(a => {
      if (target["@context"].indexOf(a) === -1) target["@context"].push(a);
    });
  } else {
    context.unshift(target["@context"]);
    target["@context"] = context;
  }
  if (count === 0) {
    log(target, "WARNING", "住所表記が見つかりませんでした");
  }

  return target;
};
